"""Functions for implementing diff cascades for re-definitions of function calls
and markdown cells.
"""
from six import string_types
def cascade(sequence, full=False):
    """Restores a sequence of string definitions using the first entry as the
    original and then applying a series of :func:`~acorn.logging.diff.restore`
    calls.

    Args:
        sequence (list): of results returned by
          :func:`~acorn.logging.diff.compress`, except that the first entry should
          be a list of string entries for the very first instance.
        full (bool): when True, return all the intermediate entries as well;
          otherwise they are not stored in memory and only the final entry in
          the list is returned.
    """
    if len(sequence) == 1:
        return sequence[0]
    
    left = sequence[0]
    if full:
        intermed = []
    for cdiff in sequence[1:]:
        right = restore(cdiff, left)
        if full:
            intermed.append(right)
        left = right

    return left

def restore(cdiff, a):
    """Restores the full text of either the edited text using the
    compressed diff.

    Args:
        cdiff (dict): compressed diff returned by
          :func:`~acorn.logging.diff.compress`. 
        a (str or list): *original* string or list of strings to use as a
          reference to restore the edited version.
    """
    left = a.splitlines(1) if isinstance(a, string_types) else a
    lrest = []
    iline = 0
    
    for i, line in enumerate(left):
        if iline not in cdiff:
            lrest.append("  " + line)
            iline += 1
        else:
            cs = [l[0] for l in cdiff[iline]]
            add = cs.count('+') - cs.count('-')
            lrest.extend(cdiff[iline])
            iline += add + 1
            
    for i in sorted(cdiff.keys()):
        if i >= len(left):
            lrest.extend(cdiff[i])

    from difflib import restore
    return list(restore(lrest, 2))
    
def compress(a, b):
    """Performs the *compressed* diff of `a` and `b` such that the original
    contents of the :func:`difflib.ndiff` call can be reconstructed using
    :func:`~acorn.logging.diff.restore`.

    Args:
        a (str or list): *original* string or list of strings to diff.
        b (str or list): *edited* string or list of strings to diff.
    """
    from difflib import ndiff
    left = a.splitlines(1) if isinstance(a, string_types) else a
    right = b.splitlines(1) if isinstance(b, string_types) else b
    ldiff = list(ndiff(left, right))
    
    result = {}
    latest = None   
    combo = None
    icombo = 0
    iorig = 0
    
    for i, line in enumerate(ldiff):
        cs = [l[0] for l in ldiff[i:min((i+4, len(ldiff)))]]
        if cs[0] != ' ':
            #Initialize a new entry in the diff list.
            if latest is None:
                latest = iorig
                result[latest] = []
                
            #We have to be careful. At a minimum, there may be a '-' or a '+' when the lines are 
            #completely added or deleted. When they are *altered*, then we also expect one or
            #more '?' lines showing the differences.
            if combo is None:
                if cs[0] == '-':
                    #Check whether the next lines have one of these combinations:
                    if (len(cs) >=3 and cs[1] == '+' and cs[2] == '?'):
                        combo = 3
                    elif (len(cs) >= 4 and cs[1] == '?' and cs[2] == '+'
                          and cs[3] == '?'):
                        combo = 4
                    else:
                        #This is a stand-alone deletion.
                        combo = 1
                elif cs[0] == '+':
                    #This is for the stand-alone addition.
                    combo = 1
                
            if icombo < combo:
                result[latest].append(line)
                icombo += 1
            
            if icombo == combo:
                if combo > 1:
                    latest = None
                combo = None
                icombo = 0
                if cs[0] != '+':
                    iorig += 1
        else:
            latest = None
            iorig += 1

    return result
    



from acorn.logging.database import set_task, set_writeable

#Add an exit handler so that in-memory collections are cleaned up correctly and
#saved to disk if the kernel is told to shut down.
import atexit
from acorn.logging.database import cleanup
atexit.register(cleanup)

import acorn.subclass
import acorn.importer

def is_ipython(): # pragma: no cover
    """Returns True if acorn is imported from an ipython notebook or terminal.
    """
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

if is_ipython(): # pragma: no cover
    from IPython import get_ipython
    ip = get_ipython()
    #Load the ipython extension for automatic decoration of new methods and
    #classes.
    ip.magic("load_ext acorn.ipython")

    #Add a javascript event handler for the markdown rendering so that we can
    #log it to the acorn database. The actual javascript is contained in
    #`./js/ipython.js`; we just read it in here and register it with ipython.
    from acorn.utility import reporoot
    from os import path
    jspath = path.join(reporoot, "js", "ipython.js")
    with open(jspath) as f:
        jscode = f.read()
    ip.run_cell_magic("javascript", "", jscode)

    #Overwrite the JPEG formatters so that plots produced with inline or
    #notebook modes will be auto-saved as images in the database.
    from acorn.ipython import AcornPNGFormatter, AcornJPEGFormatter
    formatters = ip.display_formatter.formatters
    oldpng = formatters["image/png"]
    formatters["image/png"] = AcornPNGFormatter(parent=oldpng)
    oldjpeg = formatters["image/jpeg"]
    formatters["image/jpeg"] = AcornJPEGFormatter(parent=oldjpeg)
    
import acorn.analyze



 #!/usr/bin/python
from acorn import msg
def examples():
    """Prints examples of using the script to the console using colored output.
    """
    script = "ACORN Automatic Computational Research Notebook"
    explain = ("Research involves repeated testing of hypotheses; for "
               "computational research, these \"experiments\" have a rapid "
               "turnover time so that several hundred could happen in a single "
               "day. ACORN automatically generates research notebooks from "
               "such experimentation within the context of an `ipython` "
               "notebook.\n\nThis script configures the ACORN backend.")
    contents = [(("Create a custom configuration folder for packages that "
                  "ship with ACORN."), 
                 "acrn.py configure packages",
                 "This creates a `~/.acorn` directory with copies of all the "
                 "default package configuration and descriptor files. You can "
                 "edit the configurations by opening the files from that "
                 "folder.")]
    required = ("")
    output = ("")
    details = ("")
    outputfmt = ("")

    msg.example(script, explain, contents, required, output, outputfmt, details)

script_options = {
    "commands": {"nargs": "+",
                 "help": "List of commands and sub-commands to run."},
    }
"""dict: default command-line arguments and their
    :meth:`argparse.ArgumentParser.add_argument` keyword arguments.
"""

def _parser_options():
    """Parses the options and arguments from the command line."""
    #We have two options: get some of the details from the config file,
    import argparse
    from acorn import base
    pdescr = "ACORN setup and custom configuration"
    parser = argparse.ArgumentParser(parents=[base.bparser], description=pdescr)
    for arg, options in script_options.items():
        parser.add_argument(arg, **options)
        
    args = base.exhandler(examples, parser)
    if args is None:
        return

    return args

def _conf_packages(args):
    """Runs custom configuration steps for the packages that ship with support
    in acorn.
    """
    from acorn.config import config_dir
    from os import path
    from acorn.base import testmode
    target = config_dir(True)
    alternate = path.join(path.abspath(path.expanduser("~")), ".acorn")
    if not testmode and target != alternate:# pragma: no cover
        msg.err("Could not configure custom ~/.acorn directory.")
        exit(0)

    from acorn.utility import reporoot
    from glob import glob
    from os import chdir, getcwd
    from shutil import copy
    current = getcwd()
    source = path.join(reporoot, "acorn", "config")
    chdir(source)
    count = 0

    #For the unit testing, we don't clobber the local directory, so the copies
    #are disabled.
    for json in glob("*.json"):
        if not testmode:# pragma: no cover
            copy(json, target)
        count += 1
    for cfg in glob("*.cfg"):
        if not testmode:# pragma: no cover
            copy(cfg, target)
        count += 1

    #Switch the directory back to what it was.
    chdir(current)        
    msg.okay("Copied {0:d} package files to {1}.".format(count, target))
    
def _run_configure(subcmd, args):
    """Runs the configuration step for the specified sub-command.
    """
    maps = {
        "packages": _conf_packages
        }
    if subcmd in maps:
        maps[subcmd](args)
    else:
        msg.warn("'configure' sub-command {} is not supported.".format(subcmd))

def run(args):
    """Runs the acorn setup/configuration commands.
    """
    if args is None:
        return
    
    cmd = args["commands"][0]
    if cmd == "configure":
        if len(args["commands"]) < 2:# pragma: no cover
            msg.err("'configure' command requires a second, sub-command "
                    "parameter. E.g., `acorn.py configure packages`.")
            exit(0)

        subcmd = args["commands"][1]
        _run_configure(subcmd, args)

if __name__ == '__main__': # pragma: no cover
    run(_parser_options())



"""Class for altering the search paths within acorn for new/undecorated packages
(i.e., those that weren't checked explicitly when it was first designed.
"""
from acorn import msg
_packages = {}
"""dict: keys are package names; values are `bool` indicating whether `acorn`
should decorate the package with the loggers.
"""
_special = ["scipy", "sklearn", "matplotlib"]
"""list: of packages that have special extensions or import behavior and cannot
be deco-imported automatically by `acorn`.
"""
    
def _load_package_config(reload_=False):
    """Loads the package configurations from the global `acorn.cfg` file.
    """
    global _packages
    from acorn.config import settings
    packset = settings("acorn", reload_)
    if packset.has_section("acorn.packages"):
        for package, value in packset.items("acorn.packages"):
            _packages[package] = value.strip() == "1"

def reload_cache():
    """Reloads the configuration file settings for which packages to decorate.
    """
    global _packages
    _packages = {}
    _load_package_config(True)

hooks = []
"""list: of package names that have *already* been intercepted by our
loader/decorator. This allows us to skip them the next time they are imported
(by our scripts) so that we don't get into an infinite loop.
"""
    
import sys
class AcornMetaImportFinder(object):
    """Overrides the default `import` behavior of python for packages so that we
    can intercept and decorate certain packages, but not others.

    Args:
        prefix (str): prefix on import full names before they are considered
          loadable by acorn. Also available as an attribute.
    """
    def __init__(self, prefix="acorn"):
        self.prefix = prefix
    
    def find_module(self, fullname, packpath=None):
        global hooks
        if (fullname[0:len(self.prefix)] != self.prefix):
            return None

        if fullname.count('.') > 1:
            #Acorn isn't setup to work with sub-packages.
            return None

        already = False
        package = fullname.split('.')[-1]
        if package in _packages and _packages[package]:
            if package not in hooks:
                #We can import this one and decorate it using the default or
                #custom tools in acorn.
                hooks.append(package)
                msg.okay("Import override: '{}'.".format(package), 3)
                return AcornDecoratingLoader(package)
            else:
                already = True

        if not already:
            msg.info("Skipping import override: '{}'.".format(packpath), 3)

def load_decorate(package):
    """Imports and decorates the package with the specified name.
    """
    # We import the decoration logic from acorn and then overwrite the sys.module
    # for this package with the decorated, original pandas package.
    from acorn.logging.decoration import set_decorating, decorating
    
    #Before we do any imports, we need to set that we are decorating so that
    #everything works as if `acorn` wasn't even here.
    origdecor = decorating
    set_decorating(True)

    #If we try and import the module directly, we will get stuck in a loop; at
    #some point we must invoke the built-in module loader from python. We do
    #this by removing our sys.path hook.
    import sys    
    from importlib import import_module
    apack = import_module(package)    
    from acorn.logging.decoration import decorate
    decorate(apack)

    sys.modules["acorn.{}".format(package)] = apack

    #Set the decoration back to what it was.
    from acorn.logging.decoration import set_decorating
    set_decorating(origdecor)
    
    return apack

class AcornStandardLoader(object): # pragma: no cover
    #I am having issues with intercepting all loads. At the least, pytest won't
    #import properly, so I am disabling this for now.
    """Loads packages *without* decoration, but with the correct flags enabled
    so that if those packages imported acorn decorated ones, they don't run all
    the logging machinery during imports.
    """
    def __init__(self, package):
        self.package = package
        
    def load_module(self, fullname):
        from acorn.logging.decoration import set_decorating, decorating
        global hooks
        odecor = decorating
        set_decorating(True)

        from importlib import import_module
        mod = import_module(fullname, self.package)
        hooks.remove(fullname)
        
        set_decorating(odecor)
        return mod

class AcornDecoratingLoader(object):
    """Loads packages that need to be decorated for automatic logging by
    `acorn`.

    Args:
        package (str): name of the package being loaded.
    """
    def __init__(self, package):
        self.package = package
        
    def load_module(self, fullname):
        if fullname in sys.modules: # pragma: no cover.
            msg.info("Reusing existing import for '{}'".format(fullname), 3)
            mod = sys.modules[fullname]
        else:
            msg.info("Decorating import for '{}'".format(self.package), 3)
            #First we import the package, then the specified sub-module that
            #they asked for.
            if self.package in _special:
                from importlib import import_module
                mod = import_module(fullname)
            else:
                mod = load_decorate(self.package)

        return mod
    
_load_package_config()
sys.meta_path.insert(0, AcornMetaImportFinder())

#TODO: we still need to get a package manager going. When we import the modules,
#we should always use the original package items. sklearn dies when numpy has
#already been decorated because our subclass doesn't jive with the c-extension
#modules it uses. But it is okay if it can use the original numpy class. If we
#get our own special __getattr__ method for acorn, we could return the decorated
#versions and still have the other imports work correctly. We lose the ability
#for the user to import directly from the module later (they would always have
#to go through acorn). Or perhaps we could look at the value of the global
#`decorating` and then decide which of the objects to return in getattr.



# We import the decoration logic from acorn and then overwrite the sys.module
# for this package with the decorated, original sklearn package.

from acorn.logging.decoration import set_decorating, decorating
#Before we do any imports, we need to set that we are decorating so that
#everything works as if `acorn` wasn't even here.
origdecor = decorating
set_decorating(True)

from sklearn import *
import sklearn as skl
from acorn.logging.decoration import decorate
decorate(skl)

import sys
sys.modules[__name__] = skl

import acorn.numpy as anp

#Set the decoration back to what it was.
from acorn.logging.decoration import set_decorating
set_decorating(origdecor)
